/*
Copyright (C) 2001 Chao-Kuo Lin

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "ckDialog.h"
#include "ckTraverser.h"
#include "ckControl.h"
#include "ckScanner.h"
#include "ckStream.h"
#include "ckMiscParser.h"
#include "ckValueWrapper.h"

//
// ckControlTraverser 
//
bool ckControlTraverser::traverse(ckIterator<ckControl> &it)
{
    _LSDEBUG1("ckControlTraverser() ctor");

    bool result = true;
    for(it.first(); it.isMore(); it.next())
    {
        result = _processItem(it.get());
        if(!result)
            break;
    }
    return result;
}


//
// ckWindowCreateTraverser
//
bool ckWindowCreateTraverser::_processItem(ckControl &item) 
{
    _LSDEBUGFUNS("ckWindowCreateTraverser::_processItem()");

    {//scope cleaning
        RECT r;
        DWORD style;
        DWORD xstyle;
        assert(!item.getHWND());

        ckSizeParser sizeparser(r, item);
        if(sizeparser.parse(ckStrStream(item.getAttribute(SIZE_ATTRIB))) == SUCCESS)
        {
            _LSDEBUG1("ckWindowCreateTraverser::_processItem() dumping RECT:");
            _LSDEBUG2("left : %d", r.left);
            _LSDEBUG2("top  : %d", r.top);
            _LSDEBUG2("right: %d", r.right);
            _LSDEBUG2("top  : %d", r.top);
            xstyle  = item._parseXStyle();
            style   = item._parseStyle();
            _LSDEBUG1("ckWindowCreateTraverser::_processItem() dumping style:");
            _LSDEBUG2("style : %d", style);
            _LSDEBUG2("xstyle : %d", xstyle);
            _LSDEBUG1("ckWindowCreateTraverser::_processItem() dumping stuff:");
            _LSDEBUG2("classname: %s", item.getClassName().c_str());
            _LSDEBUG2("xstyle : %s", item.getAttribute(TITLE_ATTRIB).c_str());

            item.createwindow(  xstyle,
                                item.getClassName().c_str(),
			                    item.getAttribute(TITLE_ATTRIB).c_str(),
                                style,
			                    r.left,                                     
			                    r.top,                                      
			                    r.right,                                    
			                    r.bottom,                                   
                                item.parentHWND(),                               
			                    NULL,
                                ckControl::_hInstance);

        }
    }

    _LSDEBUG2("ckWindowCreateTraverser::_processItem() window %08X", (DWORD)item.getHWND());

    if(item.getHWND())
    {
        SetWindowLong(item.getHWND(), GWL_USERDATA, (LONG)&item);
        if(ckControlTraverser::_processItem(item))
        {
            item.notify(CKINIT_EVENT); // init would be that everything's done
            return true;
        }
    }

    return false;
}



//
// ckCloneTraverser
//
ckCloneTraverser::ckCloneTraverser()
:   _specie(1)
{   
    _LSDEBUG1("ckCloneTraverser() ctor");
    _current = NULL;    
}

auto_ptr<ckControl> ckCloneTraverser::getClone() 
{ 
    assert(_specie.size() == 1);
    _LSDEBUG1("ckCloneTraverser::getClone()");
    return _specie.pop_back(); 
}

auto_vector<ckControl> ckCloneTraverser::getClones() 
{ 
    _LSDEBUG1("ckCloneTraverser::getClones()");
    return _specie; 
}

bool ckCloneTraverser::_processItem(ckControl &item) 
{
    auto_ptr<ckControl> temp(item.Clone());
    ckControl *ptemp = temp.get();
    assert(ptemp);
    
    _LSDEBUG2("ckCloneTraverser::_processItem() cloning: %s", ptemp->getAttribute(ID_ATTRIB).c_str());

    if(_current == NULL)
    {//top
        _LSDEBUG1("ckCloneTraverser::_processItem() cloned toplevel");
        _specie.push_back(temp);
        _current = ptemp;
    }
    else
    {//non top
        _LSDEBUG1("ckCloneTraverser::_processItem() cloned child");
        _current->addChild(temp);
        _current = ptemp;
    }

    _LSDEBUG1("ckCloneTraverser::_processItem() _current = child");
    
    ckControlTraverser::_processItem(item);
    
    _LSDEBUG1("ckCloneTraverser::_processItem() _current = parent");
    _current = _current->parent();

    return true;
}


//
// ckRemoveTraverser
//
ckRemoveTraverser::ckRemoveTraverser(ckControl &control)
:   _control(control)
{}

bool ckRemoveTraverser::_processItem(ckControl &item)
{
    if(&item == &_control)
    {
        ckControl *parent = item.parent();
        if(parent)
        {
            ckError error = parent->removeChild(item);
            assert(error == SUCCESS);
            return false;
        }
        return true;    //false to stop the traversal
                        //traverser's caller is responsible for checking
                        //if return is true that means nothing is removed
                        //either means the item is toplevel or doesn't exist
    }

    return ckControlTraverser::_processItem(item);
}



//
// ckReplaceEventTraverser
//
ckReplaceEventTraverser::ckReplaceEventTraverser(string const &rootDialogID)
:   _rootDialogID(rootDialogID)
{
}

bool ckReplaceEventTraverser::_processItem(ckControl &item)
{
    item.eventVariableSubstitution(_rootDialogID);
    return ckControlTraverser::_processItem(item);
}


//
// ckControlPathTraverser
//
ckControlPathTraverser::ckControlPathTraverser(ckPositionableStream &stream)
:   _stream(stream),
    _scanner(stream)
{
    _LSDEBUG1("ckControlPathTraverser() ctor");
}

bool ckControlPathTraverser::traverse(ckIterator<ckControl> &it)
{
    _LSDEBUGFUNS("ckControlPathTraverser::traverse()");

    bool result = false;
    for(it.first(); it.isMore(); it.next())
    {
        if(_test(it.get()))
        {
            _LSDEBUG1("ckControlPathTraverser::traverse() test succeeded");

            result = _processItem(it.get());
            if(result)
            {
                _LSDEBUG1("ckControlPathTraverser::traverse() found the right one");
                break;
            }
        }
    }

    _LSDEBUGFUNE("ckControlPathTraverser::traverse()");
    return result;
}


bool ckControlPathTraverser::_processItem(ckControl &item) 
{
    // traverse the child, if children failed, 
    // check against local database, then return result
    
    // we should strip the first one, since it's tested ok
    ckPositionMemento memento(_stream);
    _scanner.accept();
    if(!traverse(ckControlSingleLevelIterator(item)))
    {
        return _finalProcessing(item);
    }
    return true;
}

bool ckControlPathTraverser::_test(ckControl &item)
{
    //save the pos
    _LSDEBUGFUNS("ckControlPathTraverser::_test()");
    ckPositionMemento memento(_stream);
    if(_scanner.accept() == ckTokenString)
    {
        string str = _scanner.getString();
        return stricmp(str.c_str(), item.getAttribute(ID_ATTRIB).c_str()) == 0;
    }

    return false;
}


//
// ckGetControlTraverser 
//
ckGetControlTraverser::ckGetControlTraverser(ckPositionableStream &stream)
:   ckControlPathTraverser(stream)
{   
    _control = NULL;    
    _LSDEBUG1("ckGetControlTraverser() ctor");
}

bool ckGetControlTraverser::_finalProcessing(ckControl &item)
{
    if(_scanner.accept() == ckTokenEndStream)
    {
        _LSDEBUG1("ckGetControlTraverser::_finalProcessing() got the item");
        _control = &item;
    }
    else
    {
        _control = NULL;
    }

    return _control != NULL;
}


//
// ckControlPathNodeTraverser
//
ckControlPathNodeTraverser::ckControlPathNodeTraverser(ckPositionableStream &stream)
:   ckControlPathTraverser(stream)
{
    _LSDEBUG1("ckControlPathNodeTraverser() ctor");
}

bool ckControlPathNodeTraverser::_finalProcessing(ckControl &item)
{
    ckPositionMemento memento(_stream);
    if(_scanner.accept() == ckTokenString)
    {
        string str = _scanner.getString();
        
        _LSDEBUG2("ckControlPathNodeTraverser::_finalProcessing() got string", str.c_str());

        if(_scanner.accept() == ckTokenEndStream)
        {
            _match(item, str);
            return true;
        }
    }

    return false;
}

//
// ckAddEventTraverser
//
ckAddEventTraverser::ckAddEventTraverser(ckPositionableStream &stream, string const &action)
:   ckControlPathNodeTraverser(stream),
    _action(action)
{}

bool ckAddEventTraverser::_match(ckControl &item, string const &str)
{
    _LSDEBUG2("ckAddEventTraverser::_match() adding event", str.c_str());
    _LSDEBUG2("ckAddEventTraverser::_match() adding action", _action.c_str());

    item.addEvent(str, _action);
    return true;
}


//
// ckNotifyTraverser
//
ckNotifyTraverser::ckNotifyTraverser(ckPositionableStream &stream)
:   ckControlPathNodeTraverser(stream)
{
    _LSDEBUG1("ckNotifyTraverser() ctor");
}
    
bool ckNotifyTraverser::_match(ckControl &item, string const &str)
{
    _LSDEBUG2("ckNotifyTraverser::_match() notifying event", str.c_str());    
    item.notify(str);
    return true;
}


//
// ckSetAttributeTraverser 
//
ckSetVariableTraverser::ckSetVariableTraverser(ckPositionableStream &stream, string const &value)
:   ckControlPathNodeTraverser(stream),
    _value(value)
{
    _LSDEBUG1("ckSetVariableTraverser() ctor");
}

bool ckSetVariableTraverser::_match(ckControl &item, string const &str)
{
    _LSDEBUG2("ckSetAttributeTraverser::_match() adding attribute", str.c_str());
    item.setVariable(str, _value);
    return true;
}

//
// ckPerformTraverser 
//
ckPerformTraverser::ckPerformTraverser(ckPositionableStream &stream, ckValueWrapper &wrapper)
:   ckControlPathNodeTraverser(stream),
    _wrapper(wrapper)
{
    _LSDEBUG1("ckPerformTraverser() ctor");
}

bool ckPerformTraverser::_match(ckControl &item, string const &str)
{
    _LSDEBUG2("ckPerformTraverser::_match() performing: ", str.c_str());
    return item.performAction(str, _wrapper) == SUCCESS;
}

//
// ckRetreiveVariableTraverser 
//
ckRetreiveVariableTraverser::ckRetreiveVariableTraverser(ckPositionableStream &stream, string &result)
:   ckControlPathNodeTraverser(stream),
    _result(result)
{
    _LSDEBUG1("ckRetreiveVariableTraverser() ctor");
}

bool ckRetreiveVariableTraverser::_match(ckControl &item, string const &str)
{
    _LSDEBUG2("ckRetreiveVariableTraverser::_match() getting variable", str.c_str());
    item.retreiveVariable(str, _result);
    _LSDEBUG2("ckRetreiveVariableTraverser::_match() got variable value", _result.c_str());

    return true;
}








//
// DEBUG TRAVERSERS
//
//
// ckDebugTraverser
//
bool ckDebugTraverser::_processItem(ckControl &item)
{
    _LSDEBUG1("ckDebugTraverser::_processItem() dumping");
    _LSDEBUG2("ckDebugTraverser::_processItem() traversing item: ", item.getAttribute(ID_ATTRIB).c_str());
    _LSDEBUG2("ckDebugTraverser::_processItem() traversing title: ", item.getAttribute(TITLE_ATTRIB).c_str());
    _LSDEBUG2("ckDebugTraverser::_processItem() traversing style: ", item.getAttribute(STYLE_ATTRIB).c_str());
    _LSDEBUG2("ckDebugTraverser::_processItem() traversing xstyle: ", item.getAttribute(XSTYLE_ATTRIB).c_str());
    _LSDEBUG2("ckDebugTraverser::_processItem() traversing size: ", item.getAttribute(SIZE_ATTRIB).c_str());    
    _LSDEBUG1("ckDebugTraverser::_processItem() end dumping");

    return ckControlTraverser::_processItem(item);
}

//
// ckDebugPathTraverser 
//
ckDebugPathTraverser::ckDebugPathTraverser(ckPositionableStream &stream)
:   ckControlPathTraverser(stream)
{}

bool ckDebugPathTraverser::_finalProcessing(ckControl &item)
{
    _LSDEBUG1("ckDebugPathTraverser::_finalProcessing() dumping");
    _LSDEBUG2("ckDebugPathTraverser::_finalProcessing() traversing item: ", item.getAttribute(ID_ATTRIB).c_str());
    _LSDEBUG2("ckDebugPathTraverser::_finalProcessing() traversing title: ", item.getAttribute(TITLE_ATTRIB).c_str());
    _LSDEBUG2("ckDebugPathTraverser::_finalProcessing() traversing style: ", item.getAttribute(STYLE_ATTRIB).c_str());
    _LSDEBUG2("ckDebugPathTraverser::_finalProcessing() traversing xstyle: ", item.getAttribute(XSTYLE_ATTRIB).c_str());
    _LSDEBUG2("ckDebugPathTraverser::_finalProcessing() traversing size: ", item.getAttribute(SIZE_ATTRIB).c_str());    
    _LSDEBUG1("ckDebugPathTraverser::_finalProcessing() end dumping");

    bool result = _scanner.accept() == ckTokenEndStream;
    _LSDEBUG2("ckDebugPathTraverser::_finalProcessing() boolean code:", result);


    return result;
}


//
// ckDebugPathNodeTraverser
//
ckDebugPathNodeTraverser::ckDebugPathNodeTraverser(ckPositionableStream &stream)
:   ckControlPathNodeTraverser(stream)
{}

bool ckDebugPathNodeTraverser::_match(ckControl &item, string const &str)
{
    _LSDEBUG1("ckDebugPathNodeTraverser::_match() dumping");
    _LSDEBUG2("ckDebugPathNodeTraverser::_match() traversing item: ", item.getAttribute(ID_ATTRIB).c_str());
    _LSDEBUG2("ckDebugPathNodeTraverser::_match() traversing string: ", str.c_str());
    _LSDEBUG2("ckDebugPathNodeTraverser::_match() traversing title: ", item.getAttribute(TITLE_ATTRIB).c_str());
    _LSDEBUG2("ckDebugPathNodeTraverser::_match() traversing style: ", item.getAttribute(STYLE_ATTRIB).c_str());
    _LSDEBUG2("ckDebugPathNodeTraverser::_match() traversing xstyle: ", item.getAttribute(XSTYLE_ATTRIB).c_str());
    _LSDEBUG2("ckDebugPathNodeTraverser::_match() traversing size: ", item.getAttribute(SIZE_ATTRIB).c_str());    
    _LSDEBUG1("ckDebugPathNodeTraverser::_match() end dumping");

    return true;
}

